#pragma once

#include "StreamImpl.h"
#include <utility>

namespace stream {

template <typename T> class Stream;

// A stream which delays evaluating its input until requested to do so.
template <typename T, typename F> class DelayStreamImpl : public StreamImpl<T> {
public:
  using value_type = T;

  DelayStreamImpl(F &&f) : gen(std::forward<F>(f)) {}

  std::optional<value_type> next() override {
    if (!stream) {
      stream = gen();
    }
    return stream->next();
  }

private:
  std::optional<Stream<T>> stream;
  F gen;
};

} // namespace stream
