#pragma once

#include "DelayStreamImpl.h"
#include "EmptyStreamImpl.h"
#include "StreamImpl.h"

#include <cstdint>
#include <memory>
#include <utility>

namespace stream {

template <typename T> class Stream {
public:
  using value_type = T;

  Stream(const std::shared_ptr<StreamImpl<T>> &i) : impl(i) {}

  std::optional<T> next() { return impl->next(); }

private:
  std::shared_ptr<StreamImpl<T>> impl;
};

template <typename T> Stream<T> empty() {
  return Stream<T>(std::make_shared<EmptyStreamImpl<T>>());
}

template <typename F> decltype(std::declval<F>()()) delay(F &&func) {
  using V = decltype(std::declval<F>()());
  using T = typename V::value_type;
  return V(std::make_shared<DelayStreamImpl<T, F>>(std::forward<F>(func)));
}

// You need to implement all the funtions listed below. Read the handout for
// more information Here is a checklist:
// - once()
// - chain()
// - take()
// - filter()
// - map()
// - prime()
// - hamming()
// - pi()
// - repeat()
// - counter()

template <typename T> Stream<T> once(T val) {
  // Fill in the implementation here
  return empty<T>();
}

template <typename T>
Stream<T> chain(const Stream<T> &first, const Stream<T> &second) {
  // Fill in the implementation here
  return empty<T>();
}

template <typename T> Stream<T> repeat(T val) {
  // Fill in the implementation here
  return empty<T>();
}

/* Note: C++ sometimes has trouble working out the types of integer literals,
for example, auto x = '3' will always infer an int, even if the value is used
as an unsigned later. Integer literal notation may be helpful here, e.g.
`3ll` is always a long long, and `3ul` is always an unsigned long, aka
size_t */
Stream<size_t> counter(size_t val) {
  // Fill in the implementation here
  return empty<size_t>();
}

template <typename T> Stream<T> take(const Stream<T> &stream, std::size_t n) {
  // Fill in the implementation here
  return empty<T>();
}

template <typename T, typename F>
Stream<T> filter(const Stream<T> &stream, F &&f) {
  // Fill in the implementation here
  return empty<T>();
}

template <typename S, typename F>
Stream<decltype(std::declval<F>()(std::declval<S>()))>
map(const Stream<S> &stream, F &&f) {
  // Fill in the implementation here
  using T = decltype(std::declval<F>()(std::declval<S>()));
  return empty<T>();
}

// The prime() function returns a stream of prime numbers, in increasing order
// A prime number is a natural number greater than 1 that has no positive
// divisors other than 1 and itself
// The first 20 prime numbers are
// { 2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59, 61, 67, 71
// }
Stream<std::size_t> prime() {
  // Fill in the implementation here
  return empty<std::size_t>();
}

// (THIS IS A HELPER FUNCTION FOR hamming())
// The mergeUnique function takes two streams s0 and s1 as parameters
// Both s0 and s1 are sequences of increasing numbers (potentially infinite)
// Return a new stream of increasing numbers that contains all numbers from
// s0 and s1, removing duplicates
// e.g. if s0 = { 4, 6, 8, 10 } and s1 = { 3, 6, 9, 12, 15 },
// the returning stream should be { 3, 4, 6, 8, 9, 10, 12, 15 }
Stream<std::size_t> mergeUnique(Stream<std::size_t> s0,
                                Stream<std::size_t> s1) {
  // Fill in the implementation here
  return empty<std::size_t>();
}

// The hamming() function returns a stream of hamming numbers, in increasing
// order A hamming sequence enumerate, in ascending order with no repetitions,
// all positive integers with no prime factors other than 2, 3, or 5 The first
// 20 hamming numbers are { 1, 2, 3, 4, 5, 6, 8, 9, 10, 12, 15, 16, 18, 20, 24,
// 25, 27, 30, 32, 36 }
Stream<std::size_t> hamming() {
  // Fill in the implementation here
  return empty<std::size_t>();
}

// (THIS IS A HELPER FUNCTION FOR pi())
// The partialSum function takes a streams s of doubles (potentially infinite)
// and returns a new stream representing the partial sum of the input stream
// e.g. if s = { 1, 2, 3, 4, 5 }, the returning stream should be { 1, 3, 6, 10,
// 15 }
Stream<double> partialSum() {
  // Fill in the implementation here
  return empty<double>();
}

// The pi() function returns a stream of increasingly accurate approximations of
// pi. Use the expansion series for pi/4 for the approximation: pi/4 = 1 - 1/3 +
// 1/5 - 1/7 + 1/9 - ...
Stream<double> pi() {
  // Fill in the implementation here
  return empty<double>();
}

} // namespace stream
